#include "colormap.h"
#include "vectors.h"
#include "cppcolormap.h"

namespace {
  std::array<float, 3 * 256> black_body_data();
  std::array<float, 3 * 256> gray_colormapdata();
  std::array<float, 3 * 256> cmocean_phasemapdata();
  std::array<float, 3 * 256> diff_colormapdata(float start = 0, float end = 1);
  std::array<float, 3 * 256> hsv_colormapdata();

  std::array<float, 3 * 256> reverse_colormap(const std::array<float, 3 * 256> &data);

  std::array<float, 3 * 256> cppcolormap_to_std(const std::string& name) {
    std::array<float, 3 * 256> out{};
    auto data = cppcolormap::colormap(name);
    std::copy(data.begin(), data.end(), out.begin());
    return out;
  }
}  // namespace

std::array<float, 3 * 256> get_colormapdata(ColorMap cmap) {
  switch (cmap) {
    case ColorMap::GRAY:
      return gray_colormapdata();
    case ColorMap::HSV:
      return hsv_colormapdata();
    case ColorMap::BLACKBODY:
      return black_body_data();
    case ColorMap::VIRIDIS:
      return cppcolormap_to_std("viridis");
    case ColorMap::PRGn:
      return diff_colormapdata();
    case ColorMap::PRGn_POS:
      return diff_colormapdata(0.5, 1);
    case ColorMap::PRGn_NEG:
      return diff_colormapdata(0, 0.5);
    case ColorMap::RdBu:
      return cppcolormap_to_std("RdBu");
    default:
      throw std::logic_error("Unkown colormap!");
  }
}

/*
 * Implementation
 */

namespace {
  namespace PRGn {
    float colormap_red(float x) {
      if (x < 0.09963276982307434) {
        return 5.56064615384614E+02 * x + 6.34200000000000E+01;
      } else if (x < 0.4070911109447479) {
        return ((-1.64134573262743E+03 * x + 1.26126075878571E+03) * x + 8.30228593437549E+01) * x +
               9.96536529647110E+01;
      } else if (x < 0.5013306438922882) {
        return 1.64123076923114E+02 * x + 1.64926153846145E+02;
      } else if (x < 0.9049346148967743) {
        return ((((-4.17783076764345E+04 * x + 1.55735420068591E+05) * x - 2.27018068370619E+05) *
                     x +
                 1.61149115838926E+05) *
                    x -
                5.60588504546212E+04) *
                   x +
               7.93919652573346E+03;
      } else {
        return -2.67676923076906E+02 * x + 2.68590769230752E+02;
      }
    }
    float colormap_green(float x) {
      if (x < 0.1011035144329071) {
        return 4.30627692307691E+02 * x - 1.56923076923038E-01;
      } else if (x < 0.5013851821422577) {
        return ((2.21240993583109E+02 * x - 7.23499016773187E+02) * x + 8.74292145995924E+02) * x -
               3.78460594811949E+01;
      } else {
        return ((((-8.82260255008935E+03 * x + 3.69735516719018E+04) * x - 5.94940784200438E+04) *
                     x +
                 4.54236515662453E+04) *
                    x -
                1.66043372157228E+04) *
                   x +
               2.59449114260420E+03;
      }
    }
    float colormap_blue(float x) {
      if (x < 0.50031378865242) {
        return ((((1.32543265346288E+04 * x - 1.82876583834065E+04) * x + 9.17087085537958E+03) * x -
                 2.45909850441496E+03) *
                    x +
                7.42893247681885E+02) *
                   x +
               7.26668497072812E+01;
      } else if (x < 0.609173446893692) {
        return -3.50141636141726E+02 * x + 4.22147741147797E+02;
      } else {
        return ((1.79776073728688E+03 * x - 3.80142452792079E+03) * x + 2.10214624189039E+03) * x -
               6.74426111651015E+01;
      }
    }

    Vec3f colormap(float x) {
      auto r = colormap_red(x) / 255.f;
      auto g = colormap_green(x) / 255.f;
      auto b = colormap_blue(x) / 255.f;
      return {r, g, b};
    }
  }  // namespace PRGn

  std::array<float, 3 * 256> reverse_colormap(const std::array<float, 3 * 256> &data) {
    std::array<float, 3 * 256> reversed = {};
    for (int i = 0; i < 256; i++) {
      reversed[i * 3 + 0] = data[(255 - i) * 3 + 0];
      reversed[i * 3 + 1] = data[(255 - i) * 3 + 1];
      reversed[i * 3 + 2] = data[(255 - i) * 3 + 2];
    }
    return reversed;
  }

  std::array<float, 3 * 256> diff_colormapdata(float start, float end) {
    std::array<float, 3 * 256> data = {};
    for (int i = 0; i < 256; i++) {
      float x = static_cast<float>(i) / 255.f;
      x = x * (end - start) + start;
      Vec3f c = PRGn::colormap(x);
      data[i * 3 + 0] = c[0];
      data[i * 3 + 1] = c[1];
      data[i * 3 + 2] = c[2];
    }
    return data;
  }

  namespace HSV {
    float colormap_red(float x) {
      if (x < 0.5) {
        return -6.0 * x + 67.0 / 32.0;
      } else {
        return 6.0 * x - 79.0 / 16.0;
      }
    }
    float colormap_green(float x) {
      if (x < 0.4) {
        return 6.0 * x - 3.0 / 32.0;
      } else {
        return -6.0 * x + 79.0 / 16.0;
      }
    }
    float colormap_blue(float x) {
      if (x < 0.7) {
        return 6.0 * x - 67.0 / 32.0;
      } else {
        return -6.0 * x + 195.0 / 32.0;
      }
    }
    Vec3f colormap(float x) {
      return {colormap_red(x), colormap_green(x), colormap_blue(x)};
    }
  }  // namespace HSV

  std::array<float, 3 * 256> hsv_colormapdata() {
    std::array<float, 3 * 256> data = {};
    for (int i = 0; i < 256; i++) {
      float x         = static_cast<float>(i) / 255.f;
      auto c          = HSV::colormap(x);
      data[i * 3 + 0] = c[0];
      data[i * 3 + 1] = c[1];
      data[i * 3 + 2] = c[2];
    }
    return data;
  }

  std::array<float, 3 * 256> black_body_data() {
    return {0,
            0,
            0,
            0.0124985675549,
            0.00376710208302,
            0.00210713336054,
            0.0249971351098,
            0.00753420416604,
            0.00421426672107,
            0.0374957026647,
            0.0113013062491,
            0.00632140008161,
            0.0492581088139,
            0.0150684083321,
            0.00842853344215,
            0.0594165938721,
            0.0188355104151,
            0.0105356668027,
            0.0684471877541,
            0.0226026124981,
            0.0126428001632,
            0.0766363089899,
            0.0263697145811,
            0.0147499335238,
            0.0841678787653,
            0.0301368166641,
            0.0168570668843,
            0.0911680855116,
            0.0339039187472,
            0.0189642002448,
            0.0977278294789,
            0.0376710208302,
            0.0210713336054,
            0.103915087694,
            0.0414146679499,
            0.0231784669659,
            0.109782226099,
            0.044974293467,
            0.0252856003264,
            0.115370570052,
            0.0483647728382,
            0.027392733687,
            0.120713391496,
            0.0516062923368,
            0.0294998670475,
            0.125837934386,
            0.054715381023,
            0.0316070004081,
            0.130854584645,
            0.0576505684502,
            0.0337193392032,
            0.13604190193,
            0.06025908266,
            0.0358534453625,
            0.14140417663,
            0.0625542306901,
            0.0380109209821,
            0.146927404384,
            0.0645557290717,
            0.0401922566139,
            0.152598730721,
            0.0662787359665,
            0.0423389683918,
            0.158389635306,
            0.0677778384763,
            0.0444339133373,
            0.164222599234,
            0.0692401824912,
            0.046456344033,
            0.170089016456,
            0.0706848851824,
            0.0484086974279,
            0.175987920545,
            0.0721120614255,
            0.0502958207167,
            0.181918421843,
            0.0735218171731,
            0.0521219928396,
            0.187879698899,
            0.0749142499447,
            0.0538910122917,
            0.19387099103,
            0.0762894492712,
            0.0556062682658,
            0.199891591854,
            0.0776474970992,
            0.0572707988285,
            0.205940843628,
            0.0789884681608,
            0.0588873389018,
            0.212018132302,
            0.0803124303088,
            0.0604583601449,
            0.218122883178,
            0.0816194448242,
            0.0619861043454,
            0.224254557086,
            0.0829095666958,
            0.0634726115611,
            0.230412647017,
            0.0841828448758,
            0.0649197439858,
            0.236596675144,
            0.0854393225134,
            0.0663292063043,
            0.242806190184,
            0.0866790371676,
            0.0677025631472,
            0.249040765054,
            0.0879020210012,
            0.0690412541328,
            0.255299994784,
            0.0891083009582,
            0.0703466068909,
            0.26158349466,
            0.0902978989252,
            0.0716198483882,
            0.267890898557,
            0.0914708318774,
            0.0728621148187,
            0.274221857454,
            0.0926271120122,
            0.0740744602728,
            0.280576038088,
            0.0937667468691,
            0.0752578643655,
            0.286953121761,
            0.0948897394388,
            0.07641323897,
            0.293352803248,
            0.0959960882604,
            0.077541434183,
            0.299774789817,
            0.0970857875089,
            0.0786432436236,
            0.306218800348,
            0.0981588270728,
            0.0797194091565,
            0.312684564525,
            0.0992151926224,
            0.0807706251109,
            0.319171134912,
            0.10025502621,
            0.0818055914811,
            0.325676914545,
            0.101278631704,
            0.0828412386963,
            0.332201791044,
            0.102285972873,
            0.0838775716485,
            0.33874565084,
            0.103277010196,
            0.0849145970685,
            0.345308379611,
            0.104251700818,
            0.0859523214123,
            0.351889862505,
            0.105209998536,
            0.0869907508714,
            0.358489984344,
            0.106151853789,
            0.0880298913836,
            0.365108629799,
            0.107077213638,
            0.0890697486419,
            0.371745683545,
            0.107986021747,
            0.0901103281045,
            0.378401030403,
            0.108878218356,
            0.0911516350032,
            0.385074555462,
            0.109753740249,
            0.0921936743521,
            0.391766144185,
            0.110612520723,
            0.0932364509556,
            0.398475682511,
            0.111454489549,
            0.0942799694164,
            0.405203056934,
            0.112279572927,
            0.0953242341425,
            0.411948154577,
            0.113087693439,
            0.0963692493548,
            0.418710863264,
            0.113878770002,
            0.0974150190938,
            0.425491071569,
            0.114652717803,
            0.098461547226,
            0.432288668871,
            0.115409448248,
            0.0995088374503,
            0.439103545394,
            0.116148868889,
            0.100556893304,
            0.445935592246,
            0.116870883353,
            0.101605718168,
            0.452784701449,
            0.117575391273,
            0.102655315275,
            0.459650765971,
            0.118262288197,
            0.10370568771,
            0.46653367974,
            0.118931465509,
            0.10475683842,
            0.473433337672,
            0.11958281033,
            0.105808770216,
            0.480349635682,
            0.120216205427,
            0.10686148578,
            0.487282470697,
            0.120831529099,
            0.107914987668,
            0.494231740665,
            0.121428655073,
            0.108969278312,
            0.501197344567,
            0.122007452382,
            0.110024360032,
            0.508179182414,
            0.122567785241,
            0.11108023503,
            0.515177155256,
            0.123109512914,
            0.112136905403,
            0.522191165184,
            0.123632489573,
            0.113194373138,
            0.529221115323,
            0.12413656415,
            0.114252640123,
            0.536266909839,
            0.12462158018,
            0.115311708147,
            0.543328453928,
            0.125087375633,
            0.116371578905,
            0.550405653818,
            0.125533782739,
            0.117432253997,
            0.557498416761,
            0.125960627799,
            0.118493734936,
            0.564606651025,
            0.126367730992,
            0.11955602315,
            0.571730265893,
            0.126754906165,
            0.120619119982,
            0.57886917165,
            0.12712196061,
            0.121683026697,
            0.586023279576,
            0.127468694834,
            0.122747744479,
            0.59319250194,
            0.127794902308,
            0.123813274439,
            0.600376751988,
            0.12810036921,
            0.124879617617,
            0.607575943936,
            0.128384874145,
            0.125946774979,
            0.614789992957,
            0.128648187853,
            0.127014747427,
            0.622018815174,
            0.128890072894,
            0.128083535793,
            0.629262327647,
            0.129110283326,
            0.12915314085,
            0.636520448365,
            0.129308564349,
            0.130223563305,
            0.643793096232,
            0.129484651934,
            0.131294803809,
            0.65108019106,
            0.129638272432,
            0.132366862954,
            0.658381653554,
            0.129769142155,
            0.133439741276,
            0.665697405303,
            0.129876966926,
            0.134513439256,
            0.67302736877,
            0.129961441609,
            0.135587957327,
            0.680371467279,
            0.130022249607,
            0.136663295866,
            0.686819633179,
            0.132061139735,
            0.137260535772,
            0.690577321338,
            0.139841365035,
            0.136421938331,
            0.694333453232,
            0.147366583596,
            0.135554156099,
            0.69808804669,
            0.154670100683,
            0.134656431336,
            0.701841119323,
            0.161779110231,
            0.133727968564,
            0.705592688522,
            0.16871613022,
            0.132767931803,
            0.709342771462,
            0.175500034803,
            0.131775441528,
            0.713091385109,
            0.182146813036,
            0.130749571337,
            0.716838546221,
            0.188670137478,
            0.129689344281,
            0.720584271352,
            0.195081797593,
            0.128593728817,
            0.724328576854,
            0.201392035109,
            0.127461634333,
            0.728071478883,
            0.207609807026,
            0.126291906198,
            0.7318129934,
            0.213742994383,
            0.125083320255,
            0.735553136174,
            0.219798569795,
            0.123834576701,
            0.739291922788,
            0.225782733244,
            0.122544293247,
            0.743029368637,
            0.231701023129,
            0.121210997469,
            0.746765488936,
            0.237558407857,
            0.11983311822,
            0.75050029872,
            0.243359361942,
            0.118408975956,
            0.754233812848,
            0.249107929694,
            0.116936771817,
            0.757966046005,
            0.254807778863,
            0.115414575242,
            0.761697012704,
            0.260462246095,
            0.113840309888,
            0.765426727292,
            0.266074375673,
            0.112211737544,
            0.769155203948,
            0.271646952701,
            0.110526439697,
            0.77288245669,
            0.277182531676,
            0.108781796299,
            0.776608499373,
            0.28268346119,
            0.106974961217,
            0.780333345697,
            0.288151905399,
            0.105102833683,
            0.784057009203,
            0.293589862731,
            0.103162024955,
            0.787779503281,
            0.298999182276,
            0.101148819133,
            0.791500841169,
            0.304381578185,
            0.0990591268518,
            0.795221035957,
            0.309738642359,
            0.0968884301939,
            0.798940100587,
            0.315071855678,
            0.0946317167004,
            0.802658047857,
            0.320382597962,
            0.0922833997264,
            0.806374890426,
            0.325672156837,
            0.0898372215234,
            0.810090640808,
            0.330941735639,
            0.0872861342463,
            0.813805311384,
            0.336192460496,
            0.0846221524179,
            0.817518914394,
            0.341425386671,
            0.0818361680288,
            0.821231461948,
            0.346641504267,
            0.0789177160322,
            0.824942966022,
            0.351841743372,
            0.0758546729705,
            0.828653438464,
            0.357026978703,
            0.0726328638943,
            0.832362890991,
            0.362198033804,
            0.0692355410618,
            0.836071335196,
            0.367355684865,
            0.065642679414,
            0.839778782546,
            0.372500664176,
            0.0618300036213,
            0.843485244386,
            0.377633663287,
            0.0577676104278,
            0.847190731941,
            0.382755335881,
            0.0534179600786,
            0.850895256314,
            0.387866300395,
            0.0487328445222,
            0.854598828494,
            0.39296714243,
            0.0436486154689,
            0.858301459351,
            0.398058416948,
            0.0380796196638,
            0.862003159643,
            0.403140650294,
            0.0322904386835,
            0.865703940014,
            0.408214342055,
            0.0264118339446,
            0.869403810997,
            0.413279966769,
            0.0204432589796,
            0.870620849202,
            0.420259688127,
            0.0196207043343,
            0.871622154301,
            0.427341619388,
            0.0192250591085,
            0.87260643877,
            0.434372519047,
            0.0188396073492,
            0.87357362974,
            0.441355114094,
            0.0184645577569,
            0.874523653517,
            0.448291940683,
            0.0181001190322,
            0.875456435563,
            0.45518536159,
            0.0177464998757,
            0.876371900482,
            0.462037581704,
            0.017403908988,
            0.877269972004,
            0.468850661816,
            0.0170725550697,
            0.878150572968,
            0.47562653091,
            0.0167526468212,
            0.879013625306,
            0.482366997164,
            0.0164443929432,
            0.879859050028,
            0.489073757822,
            0.0161480021363,
            0.880686767202,
            0.495748408058,
            0.0158636831011,
            0.881496695939,
            0.502392448976,
            0.015591644538,
            0.882288754373,
            0.509007294828,
            0.0153320951477,
            0.883062859646,
            0.51559427955,
            0.0150852436308,
            0.883818927886,
            0.522154662692,
            0.0148512986878,
            0.884556874192,
            0.528689634804,
            0.0146304690194,
            0.885276612612,
            0.535200322339,
            0.014422963326,
            0.885978056125,
            0.541687792133,
            0.0142289903082,
            0.886661116621,
            0.548153055488,
            0.0140487586667,
            0.887325704879,
            0.554597071914,
            0.013882477102,
            0.88797173055,
            0.561020752565,
            0.0137303543147,
            0.888599102131,
            0.56742496338,
            0.0135925990053,
            0.889207726947,
            0.573810527989,
            0.0134694198745,
            0.889797511128,
            0.580178230371,
            0.0133610256227,
            0.890368359587,
            0.586528817319,
            0.0132676249506,
            0.890920175994,
            0.592863000709,
            0.0131894265588,
            0.891452862757,
            0.5991814596,
            0.0131266391478,
            0.891966320994,
            0.605484842178,
            0.0130794714182,
            0.892460450511,
            0.611773767557,
            0.0130481320706,
            0.892935149775,
            0.618048827454,
            0.0130328298055,
            0.893390315889,
            0.624310587737,
            0.0130337733235,
            0.893825844568,
            0.630559589875,
            0.0130511713252,
            0.894241630107,
            0.636796352279,
            0.0130852325112,
            0.894637565357,
            0.643021371558,
            0.013136165582,
            0.895013541696,
            0.649235123688,
            0.0132041792383,
            0.895369449001,
            0.655438065104,
            0.0132894821805,
            0.895705175617,
            0.661630633721,
            0.0133922831093,
            0.896020608327,
            0.66781324989,
            0.0135127907252,
            0.89631563232,
            0.673986317291,
            0.0136512137289,
            0.896590131162,
            0.680150223773,
            0.0138077608208,
            0.896843986762,
            0.686305342141,
            0.0139826407016,
            0.897077079337,
            0.692452030892,
            0.0141760620719,
            0.897289287378,
            0.698590634912,
            0.0143882336321,
            0.89748048762,
            0.704721486124,
            0.0146193640829,
            0.897650554997,
            0.710844904108,
            0.0148696621249,
            0.897799362612,
            0.716961196674,
            0.0151393364586,
            0.897926781699,
            0.72307066041,
            0.0154285957846,
            0.89803268158,
            0.729173581195,
            0.0157376488035,
            0.898116929626,
            0.735270234683,
            0.0160667042159,
            0.89817939122,
            0.741360886763,
            0.0164159707222,
            0.89821992971,
            0.747445793991,
            0.0167856570232,
            0.89823840637,
            0.753525204,
            0.0171759718194,
            0.898234680352,
            0.759599355884,
            0.0175871238113,
            0.898208608643,
            0.76566848057,
            0.0180193216995,
            0.898160046015,
            0.771732801162,
            0.0184727741846,
            0.898088844982,
            0.777792533271,
            0.0189476899672,
            0.897994855746,
            0.783847885327,
            0.0194442777479,
            0.897877926148,
            0.789899058877,
            0.0199627462272,
            0.897737901615,
            0.795946248863,
            0.0205033041056,
            0.897574625107,
            0.801989643893,
            0.0210661600838,
            0.897387937062,
            0.808029426494,
            0.0216515228624,
            0.897177675337,
            0.814065773352,
            0.0222596011419,
            0.896943675152,
            0.820098855544,
            0.0228906036229,
            0.896685769028,
            0.826128838754,
            0.023544739006,
            0.898553007849,
            0.831233249825,
            0.0682610078157,
            0.902890379229,
            0.835257282572,
            0.127173791555,
            0.907153202363,
            0.839290415225,
            0.16936596405,
            0.911340063464,
            0.843332745276,
            0.204553625577,
            0.915449563182,
            0.847384366892,
            0.235764224808,
            0.919480314548,
            0.851445370991,
            0.264391329421,
            0.923430940957,
            0.85551584532,
            0.291204862021,
            0.927300074214,
            0.859595874528,
            0.316679372942,
            0.931086352617,
            0.863685540241,
            0.341129701928,
            0.934788419075,
            0.867784921127,
            0.364776124263,
            0.938404919262,
            0.871894092973,
            0.387779074121,
            0.941934499804,
            0.876013128743,
            0.410259138592,
            0.945375806484,
            0.880142098652,
            0.43230927905,
            0.948727482467,
            0.884281070227,
            0.454002666231,
            0.951988166539,
            0.888430108369,
            0.475397903,
            0.955156491362,
            0.892589275414,
            0.496542620799,
            0.958231081722,
            0.896758631197,
            0.51747602533,
            0.96121055279,
            0.900938233107,
            0.538230741549,
            0.964093508368,
            0.905128136146,
            0.55883417852,
            0.966878539132,
            0.909328392984,
            0.579309557384,
            0.969564220862,
            0.913539054016,
            0.599676697971,
            0.972149112643,
            0.917760167413,
            0.619952629264,
            0.974631755056,
            0.921991779176,
            0.640152069209,
            0.977010668325,
            0.92623393319,
            0.660287806118,
            0.979284350439,
            0.930486671268,
            0.680371005012,
            0.981451275229,
            0.934750033206,
            0.700411455972,
            0.9835098904,
            0.939024056828,
            0.720417777213,
            0.985458615506,
            0.943308778035,
            0.740397582441,
            0.987295839871,
            0.94760423085,
            0.760357619768,
            0.989019920437,
            0.951910447464,
            0.780303887791,
            0.990629179549,
            0.956227458278,
            0.80024173319,
            0.992121902645,
            0.960555291948,
            0.820175933243,
            0.993496335869,
            0.964893975428,
            0.840110765987,
            0.994750683575,
            0.969243534009,
            0.860050070139,
            0.995883105733,
            0.973603991364,
            0.879997296537,
            0.99689171521,
            0.97797536958,
            0.899955552467,
            0.997774574927,
            0.982357689206,
            0.919927640041,
            0.998529694871,
            0.986750969284,
            0.939916089534,
            0.999155028965,
            0.991155227389,
            0.959923188443,
            0.999648471753,
            0.995570479665,
            0.979951006929,
            1,
            1,
            1};
  }
  std::array<float, 3 * 256> gray_colormapdata() {
    // https://peterkovesi.com/projects/colourmaps/index.html
    return {
        0,        0,        0,        0.0055606, 0.0055631, 0.0055625, 0.011212, 0.011219, 0.011217,
        0.016877, 0.016885, 0.016883, 0.022438,  0.022448,  0.022445,  0.027998, 0.028011, 0.028008,
        0.03354,  0.033554, 0.033551, 0.039316,  0.039333,  0.039329,  0.0447,   0.044719, 0.044714,
        0.049695, 0.049713, 0.049709, 0.054322,  0.054343,  0.054338,  0.058797, 0.058821, 0.058815,
        0.063041, 0.063062, 0.063057, 0.066996,  0.067019,  0.067014,  0.070827, 0.070853, 0.070846,
        0.074487, 0.074511, 0.074505, 0.077967,  0.077992,  0.077986,  0.081449, 0.081477, 0.08147,
        0.084728, 0.084752, 0.084746, 0.087913,  0.087939,  0.087933,  0.091003, 0.091031, 0.091024,
        0.094004, 0.094034, 0.094027, 0.096981,  0.097012,  0.097005,  0.10004,  0.10007,  0.10006,
        0.10311,  0.10314,  0.10314,  0.10618,   0.10621,   0.1062,    0.10924,  0.10928,  0.10927,
        0.11232,  0.11235,  0.11234,  0.11538,   0.11541,   0.1154,    0.11852,  0.11855,  0.11854,
        0.12157,  0.1216,   0.12159,  0.12471,   0.12474,   0.12473,   0.12787,  0.1279,   0.12789,
        0.13102,  0.13105,  0.13105,  0.13418,   0.13421,   0.1342,    0.13734,  0.13738,  0.13737,
        0.14051,  0.14055,  0.14054,  0.1437,    0.14373,   0.14372,   0.14687,  0.14691,  0.1469,
        0.15011,  0.15015,  0.15014,  0.15329,   0.15333,   0.15332,   0.15656,  0.15659,  0.15659,
        0.15975,  0.15979,  0.15978,  0.16304,   0.16308,   0.16307,   0.16628,  0.16632,  0.16631,
        0.16957,  0.16961,  0.1696,   0.17281,   0.17286,   0.17285,   0.17608,  0.17612,  0.17611,
        0.1794,   0.17944,  0.17943,  0.18269,   0.18273,   0.18272,   0.18604,  0.18608,  0.18607,
        0.18934,  0.18939,  0.18938,  0.19267,   0.19272,   0.19271,   0.19602,  0.19607,  0.19606,
        0.19934,  0.19939,  0.19938,  0.20271,   0.20276,   0.20275,   0.20611,  0.20616,  0.20615,
        0.20946,  0.20951,  0.2095,   0.21284,   0.21289,   0.21288,   0.21626,  0.21632,  0.2163,
        0.21965,  0.2197,   0.21969,  0.22307,   0.22312,   0.22311,   0.22649,  0.22654,  0.22653,
        0.22988,  0.22994,  0.22992,  0.23332,   0.23337,   0.23336,   0.2368,   0.23685,  0.23684,
        0.24022,  0.24027,  0.24026,  0.24368,   0.24373,   0.24372,   0.24716,  0.24722,  0.2472,
        0.25063,  0.25069,  0.25067,  0.25412,   0.25418,   0.25417,   0.2576,   0.25766,  0.25765,
        0.26111,  0.26117,  0.26115,  0.26461,   0.26467,   0.26465,   0.26812,  0.26818,  0.26817,
        0.27164,  0.2717,   0.27168,  0.27518,   0.27525,   0.27523,   0.27872,  0.27879,  0.27877,
        0.28227,  0.28234,  0.28232,  0.2858,    0.28587,   0.28585,   0.28938,  0.28945,  0.28943,
        0.29294,  0.293,    0.29298,  0.29652,   0.29659,   0.29657,   0.30009,  0.30016,  0.30014,
        0.30368,  0.30375,  0.30373,  0.30728,   0.30734,   0.30733,   0.3109,   0.31096,  0.31095,
        0.31449,  0.31456,  0.31454,  0.31811,   0.31819,   0.31817,   0.32173,  0.32181,  0.32179,
        0.32536,  0.32543,  0.32542,  0.32902,   0.32909,   0.32907,   0.33267,  0.33274,  0.33272,
        0.3363,   0.33638,  0.33636,  0.33997,   0.34005,   0.34003,   0.34363,  0.3437,   0.34369,
        0.34731,  0.34739,  0.34737,  0.35099,   0.35107,   0.35105,   0.35468,  0.35476,  0.35474,
        0.35836,  0.35844,  0.35842,  0.36207,   0.36214,   0.36212,   0.36577,  0.36585,  0.36583,
        0.36948,  0.36956,  0.36954,  0.37319,   0.37328,   0.37326,   0.37693,  0.37701,  0.37699,
        0.38066,  0.38074,  0.38072,  0.38439,   0.38447,   0.38445,   0.38814,  0.38822,  0.3882,
        0.39189,  0.39197,  0.39195,  0.39565,   0.39574,   0.39572,   0.3994,   0.39949,  0.39946,
        0.40318,  0.40327,  0.40325,  0.40695,   0.40704,   0.40702,   0.41074,  0.41083,  0.41081,
        0.41452,  0.41461,  0.41459,  0.41831,   0.4184,    0.41838,   0.42211,  0.4222,   0.42218,
        0.42592,  0.42601,  0.42599,  0.42974,   0.42983,   0.42981,   0.43355,  0.43364,  0.43362,
        0.43739,  0.43748,  0.43746,  0.44121,   0.44131,   0.44128,   0.44505,  0.44515,  0.44512,
        0.4489,   0.449,    0.44897,  0.45275,   0.45284,   0.45282,   0.45659,  0.45669,  0.45667,
        0.46045,  0.46055,  0.46052,  0.46433,   0.46443,   0.46441,   0.46819,  0.46829,  0.46826,
        0.47208,  0.47218,  0.47215,  0.47596,   0.47606,   0.47603,   0.47985,  0.47995,  0.47992,
        0.48373,  0.48383,  0.48381,  0.48764,   0.48775,   0.48772,   0.49155,  0.49165,  0.49163,
        0.49545,  0.49556,  0.49553,  0.49936,   0.49947,   0.49944,   0.50329,  0.50339,  0.50337,
        0.50722,  0.50733,  0.5073,   0.51115,   0.51126,   0.51123,   0.51509,  0.5152,   0.51517,
        0.51904,  0.51915,  0.51912,  0.52299,   0.5231,    0.52307,   0.52695,  0.52706,  0.52703,
        0.5309,   0.53101,  0.53099,  0.53487,   0.53498,   0.53495,   0.53884,  0.53895,  0.53893,
        0.54282,  0.54293,  0.5429,   0.5468,    0.54691,   0.54688,   0.55078,  0.5509,   0.55087,
        0.55478,  0.55489,  0.55486,  0.55877,   0.55888,   0.55886,   0.56278,  0.5629,   0.56287,
        0.56679,  0.56691,  0.56688,  0.57079,   0.57091,   0.57088,   0.57481,  0.57493,  0.5749,
        0.57884,  0.57896,  0.57893,  0.58286,   0.58298,   0.58295,   0.5869,   0.58702,  0.58699,
        0.59094,  0.59106,  0.59103,  0.59498,   0.5951,    0.59507,   0.59903,  0.59916,  0.59913,
        0.60309,  0.60321,  0.60318,  0.60715,   0.60728,   0.60724,   0.61121,  0.61134,  0.61131,
        0.61527,  0.6154,   0.61537,  0.61936,   0.61948,   0.61945,   0.62343,  0.62356,  0.62353,
        0.62751,  0.62764,  0.62761,  0.6316,    0.63173,   0.6317,    0.63569,  0.63583,  0.63579,
        0.6398,   0.63993,  0.63989,  0.64389,   0.64402,   0.64399,   0.648,    0.64813,  0.6481,
        0.65212,  0.65225,  0.65222,  0.65624,   0.65637,   0.65634,   0.66035,  0.66049,  0.66045,
        0.66448,  0.66461,  0.66458,  0.66862,   0.66875,   0.66872,   0.67275,  0.67289,  0.67286,
        0.6769,   0.67703,  0.677,    0.68103,   0.68117,   0.68114,   0.68519,  0.68533,  0.6853,
        0.68935,  0.68949,  0.68945,  0.6935,    0.69364,   0.69361,   0.69766,  0.69781,  0.69777,
        0.70183,  0.70197,  0.70194,  0.70601,   0.70616,   0.70612,   0.71019,  0.71033,  0.7103,
        0.71437,  0.71451,  0.71448,  0.71855,   0.7187,    0.71867,   0.72274,  0.72289,  0.72286,
        0.72694,  0.72709,  0.72706,  0.73114,   0.73129,   0.73126,   0.73535,  0.7355,   0.73546,
        0.73957,  0.73972,  0.73968,  0.74378,   0.74393,   0.74389,   0.748,    0.74815,  0.74811,
        0.75222,  0.75237,  0.75233,  0.75644,   0.7566,    0.75656,   0.76068,  0.76083,  0.76079,
        0.76492,  0.76507,  0.76503,  0.76915,   0.76931,   0.76927,   0.7734,   0.77355,  0.77351,
        0.77765,  0.77781,  0.77777,  0.78191,   0.78206,   0.78202,   0.78616,  0.78632,  0.78628,
        0.79042,  0.79058,  0.79054,  0.79469,   0.79485,   0.79481,   0.79896,  0.79913,  0.79909,
        0.80323,  0.8034,   0.80336,  0.80752,   0.80768,   0.80764,   0.8118,   0.81196,  0.81192,
        0.81609,  0.81626,  0.81621,  0.82038,   0.82054,   0.8205,    0.82467,  0.82484,  0.8248,
        0.82898,  0.82915,  0.82911,  0.83328,   0.83345,   0.83341,   0.83759,  0.83776,  0.83772,
        0.84191,  0.84208,  0.84204,  0.84622,   0.84639,   0.84635,   0.85054,  0.85071,  0.85067,
        0.85487,  0.85504,  0.855,    0.8592,    0.85937,   0.85933,   0.86354,  0.86371,  0.86367,
        0.86787,  0.86805,  0.86801,  0.87222,   0.87239,   0.87235,   0.87656,  0.87673,  0.87669,
        0.88091,  0.88109,  0.88104,  0.88527,   0.88544,   0.8854,    0.88962,  0.8898,   0.88976,
        0.89399,  0.89417,  0.89412,  0.89835,   0.89853,   0.89849,   0.90272,  0.90291,  0.90286,
        0.9071,   0.90728,  0.90724,  0.91148,   0.91166,   0.91161,   0.91585,  0.91604,  0.91599,
        0.92024,  0.92043,  0.92038,  0.92463,   0.92482,   0.92477,   0.92903,  0.92921,  0.92917,
        0.93342,  0.93361,  0.93356,  0.93783,   0.93802,   0.93797,   0.94224,  0.94243,  0.94238,
        0.94664,  0.94683,  0.94679,  0.95106,   0.95125,   0.9512,    0.95548,  0.95567,  0.95562,
        0.9599,   0.96009,  0.96004,  0.96432,   0.96451,   0.96446,   0.96875,  0.96894,  0.9689,
        0.97318,  0.97338,  0.97333,  0.97762,   0.97781,   0.97777,   0.98206,  0.98226,  0.98221,
        0.98651,  0.9867,   0.98666,  0.99095,   0.99115,   0.9911,    0.9954,   0.9956,   0.99555,
        0.99986,  1,        1};
    //for (int i = 0; i < 256; i++) {
    //  cdata[3 * i + 0] = i / 255.f;
    //  cdata[3 * i + 1] = i / 255.f;
    //  cdata[3 * i + 2] = i / 255.f;
    //}
  }
  std::array<float, 3 * 256> cmocean_phasemapdata() {
    // from https://github.com/matplotlib/cmocean/blob/master/cmocean/rgb/phase.py
    return {0.65830839, 0.46993917, 0.04941288, 0.66433742, 0.4662019,  0.05766473, 0.67020869,
            0.46248014, 0.0653456,  0.67604299, 0.45869838, 0.07273174, 0.68175228, 0.45491407,
            0.07979262, 0.6874028,  0.45108417, 0.08667103, 0.6929505,  0.44723893, 0.09335869,
            0.69842619, 0.44335768, 0.09992839, 0.7038123,  0.43945328, 0.1063871,  0.70912069,
            0.43551765, 0.11277174, 0.71434524, 0.43155576, 0.11909348, 0.71949289, 0.42756272,
            0.12537606, 0.72455619, 0.4235447,  0.13162325, 0.72954895, 0.41949098, 0.13786305,
            0.73445172, 0.41541774, 0.14408039, 0.73929496, 0.41129973, 0.15032217, 0.74403834,
            0.40717158, 0.15654335, 0.74873695, 0.40298519, 0.16282282, 0.75332319, 0.39880107,
            0.16907566, 0.75788083, 0.39454245, 0.17542179, 0.7623326,  0.39028096, 0.18175915,
            0.76673205, 0.38596549, 0.18816819, 0.77105247, 0.38162141, 0.19461532, 0.77529528,
            0.37724732, 0.20110652, 0.77948666, 0.37281509, 0.2076873,  0.78358534, 0.36836772,
            0.21429736, 0.78763763, 0.363854,   0.22101648, 0.79161134, 0.35930804, 0.2277974,
            0.79550606, 0.3547299,  0.23464353, 0.79935398, 0.35007959, 0.24161832, 0.80311671,
            0.34540152, 0.24865892, 0.80681033, 0.34067452, 0.25580075, 0.8104452,  0.33588248,
            0.26307222, 0.8139968,  0.33105538, 0.27043183, 0.81747689, 0.32617526, 0.27791096,
            0.82089415, 0.32122629, 0.28553846, 0.82422713, 0.3162362,  0.29327617, 0.82747661,
            0.31120154, 0.30113388, 0.83066399, 0.30608459, 0.30917579, 0.83376307, 0.30092244,
            0.31734921, 0.83677286, 0.29571346, 0.32566199, 0.83969693, 0.29044723, 0.33413665,
            0.84253873, 0.28511151, 0.34279962, 0.84528297, 0.27972917, 0.35162078, 0.84792704,
            0.27430045, 0.36060681, 0.85046793, 0.26882624, 0.36976395, 0.85291056, 0.26328859,
            0.37913116, 0.855242,   0.25770888, 0.38868217, 0.85745673, 0.25209367, 0.39841601,
            0.85955023, 0.24644737, 0.40833625, 0.86151767, 0.24077563, 0.41844557, 0.86335392,
            0.23508521, 0.42874606, 0.86505685, 0.22937288, 0.43926008, 0.86661606, 0.22366308,
            0.44996127, 0.86802578, 0.21796785, 0.46084758, 0.86928003, 0.21230132, 0.47191554,
            0.87037274, 0.20667988, 0.48316015, 0.87129781, 0.2011224,  0.49457479, 0.87204914,
            0.19565041, 0.50615118, 0.87262076, 0.19028829, 0.51787932, 0.87300686, 0.18506334,
            0.5297475,  0.8732019,  0.18000588, 0.54174232, 0.87320066, 0.1751492,  0.55384874,
            0.87299833, 0.17052942, 0.56605016, 0.87259058, 0.16618514, 0.57832856, 0.87197361,
            0.16215698, 0.59066466, 0.87114414, 0.15848667, 0.60303881, 0.87009966, 0.15521687,
            0.61542844, 0.86883823, 0.15238892, 0.62781175, 0.86735858, 0.15004199, 0.64016651,
            0.8656601,  0.14821149, 0.65247022, 0.86374282, 0.14692762, 0.66470043, 0.86160744,
            0.14621386, 0.67683495, 0.85925523, 0.14608582, 0.68885204, 0.85668805, 0.14655046,
            0.70073065, 0.85390829, 0.14760576, 0.71245054, 0.85091881, 0.14924094, 0.7239925,
            0.84772287, 0.15143717, 0.73533849, 0.84432409, 0.15416865, 0.74647174, 0.84072639,
            0.15740403, 0.75737678, 0.83693394, 0.16110786, 0.76803952, 0.83295108, 0.16524205,
            0.77844723, 0.82878232, 0.16976729, 0.78858858, 0.82443225, 0.17464414, 0.7984536,
            0.81990551, 0.179834,   0.80803365, 0.81520674, 0.18529984, 0.8173214,  0.81034059,
            0.19100664, 0.82631073, 0.80531176, 0.1969216,  0.83499645, 0.80012467, 0.20301465,
            0.84337486, 0.79478367, 0.20925826, 0.8514432,  0.78929302, 0.21562737, 0.85919957,
            0.78365681, 0.22209936, 0.86664294, 0.77787898, 0.22865386, 0.87377308, 0.7719633,
            0.23527265, 0.88059043, 0.76591335, 0.24193947, 0.88709606, 0.7597325,  0.24863985,
            0.89329158, 0.75342394, 0.25536094, 0.89917908, 0.74699063, 0.26209137, 0.90476105,
            0.74043533, 0.2688211,  0.91004033, 0.73376055, 0.27554128, 0.91502,    0.72696862,
            0.28224415, 0.91970339, 0.7200616,  0.2889229,  0.92409395, 0.71304134, 0.29557159,
            0.92819525, 0.70590945, 0.30218508, 0.9320109,  0.69866732, 0.30875887, 0.93554451,
            0.69131609, 0.31528914, 0.93879964, 0.68385669, 0.32177259, 0.94177976, 0.6762898,
            0.32820641, 0.94448822, 0.6686159,  0.33458824, 0.94692818, 0.66083524, 0.3409161,
            0.94910264, 0.65294785, 0.34718834, 0.95101432, 0.64495358, 0.35340362, 0.95266571,
            0.63685208, 0.35956083, 0.954059,   0.62864284, 0.3656591,  0.95519608, 0.62032517,
            0.3716977,  0.95607853, 0.61189825, 0.37767607, 0.95670757, 0.60336117, 0.38359374,
            0.95708408, 0.59471291, 0.3894503,  0.95720861, 0.58595242, 0.39524541, 0.95708134,
            0.5770786,  0.40097871, 0.95670212, 0.56809041, 0.40664983, 0.95607045, 0.55898686,
            0.41225834, 0.95518556, 0.54976709, 0.41780374, 0.95404636, 0.5404304,  0.42328541,
            0.95265153, 0.53097635, 0.42870263, 0.95099953, 0.52140479, 0.43405447, 0.94908866,
            0.51171597, 0.43933988, 0.94691713, 0.50191056, 0.44455757, 0.94448311, 0.49198981,
            0.44970607, 0.94178481, 0.48195555, 0.45478367, 0.93882055, 0.47181035, 0.45978843,
            0.93558888, 0.46155756, 0.46471821, 0.93208866, 0.45119801, 0.46957218, 0.92831786,
            0.44073852, 0.47434688, 0.92427669, 0.43018722, 0.47903864, 0.9199662,  0.41955166,
            0.4836444,  0.91538759, 0.40884063, 0.48816094, 0.91054293, 0.39806421, 0.49258494,
            0.90543523, 0.38723377, 0.49691301, 0.90006852, 0.37636206, 0.50114173, 0.89444794,
            0.36546127, 0.5052684,  0.88857877, 0.35454654, 0.5092898,  0.88246819, 0.34363779,
            0.51320158, 0.87612664, 0.33275309, 0.51700082, 0.86956409, 0.32191166, 0.52068487,
            0.86279166, 0.31113372, 0.52425144, 0.85582152, 0.3004404,  0.52769862, 0.84866679,
            0.28985326, 0.53102505, 0.84134123, 0.27939616, 0.53422931, 0.83386051, 0.26909181,
            0.53731099, 0.82623984, 0.258963,   0.5402702,  0.81849475, 0.24903239, 0.54310763,
            0.8106409,  0.23932229, 0.54582448, 0.80269392, 0.22985664, 0.54842189, 0.79467122,
            0.2206551,  0.55090241, 0.78658706, 0.21173641, 0.55326901, 0.77845533, 0.20311843,
            0.55552489, 0.77028973, 0.1948172,  0.55767365, 0.76210318, 0.1868466,  0.55971922,
            0.75390763, 0.17921799, 0.56166586, 0.74571407, 0.1719422,  0.56351747, 0.73753498,
            0.16502295, 0.56527915, 0.72937754, 0.15846116, 0.566956,   0.72124819, 0.15225499,
            0.56855297, 0.71315321, 0.14639876, 0.57007506, 0.70509769, 0.14088284, 0.57152729,
            0.69708554, 0.13569366, 0.57291467, 0.68911948, 0.13081385, 0.57424211, 0.68120108,
            0.12622247, 0.57551447, 0.67333078, 0.12189539, 0.57673644, 0.66550792, 0.11780654,
            0.57791235, 0.65773233, 0.11392613, 0.5790468,  0.64999984, 0.11022348, 0.58014398,
            0.64230637, 0.10666732, 0.58120782, 0.63464733, 0.10322631, 0.58224198, 0.62701729,
            0.0998697,  0.58324982, 0.61941001, 0.09656813, 0.58423445, 0.61181853, 0.09329429,
            0.58519864, 0.60423523, 0.09002364, 0.58614483, 0.5966519,  0.08673514, 0.58707512,
            0.58905979, 0.08341199, 0.58799127, 0.58144971, 0.08004245, 0.58889466, 0.57381211,
            0.07662083, 0.58978633, 0.56613714, 0.07314852, 0.59066692, 0.55841474, 0.06963541,
            0.5915367,  0.55063471, 0.06610144, 0.59239556, 0.54278681, 0.06257861, 0.59324304,
            0.53486082, 0.05911304, 0.59407833, 0.52684614, 0.05576765, 0.5949003,  0.5187322,
            0.05262511, 0.59570732, 0.51050978, 0.04978881, 0.5964975,  0.50216936, 0.04738319,
            0.59726862, 0.49370174, 0.04555067, 0.59801813, 0.48509809, 0.04444396, 0.59874316,
            0.47635,    0.04421323, 0.59944056, 0.46744951, 0.04498918, 0.60010687, 0.45838913,
            0.04686604, 0.60073837, 0.44916187, 0.04988979, 0.60133103, 0.43976125, 0.05405573,
            0.60188055, 0.4301812,  0.05932209, 0.60238289, 0.42040543, 0.06560774, 0.60283258,
            0.41043772, 0.07281962, 0.60322442, 0.40027363, 0.08086177, 0.60355283, 0.38990941,
            0.08964366, 0.60381194, 0.37934208, 0.09908952, 0.60399554, 0.36856412, 0.10914617,
            0.60409695, 0.35755799, 0.11974119, 0.60410858, 0.34634096, 0.13082746, 0.6040228,
            0.33491416, 0.14238003, 0.60383119, 0.323267,   0.1543847,  0.60352425, 0.31138823,
            0.16679093, 0.60309301, 0.29931029, 0.17959757, 0.60252668, 0.2870237,  0.19279966,
            0.60181364, 0.27452964, 0.20634465, 0.60094466, 0.2618794,  0.22027287, 0.5999043,
            0.24904251, 0.23449833, 0.59868591, 0.23611022, 0.24904416, 0.5972746,  0.2230778,
            0.26382006, 0.59566656, 0.21004673, 0.2788104,  0.5938521,  0.19705484, 0.29391494,
            0.59183348, 0.18421621, 0.3090634,  0.58961302, 0.17161942, 0.32415577, 0.58720132,
            0.15937753, 0.3391059,  0.58461164, 0.14759012, 0.35379624, 0.58186793, 0.13637734,
            0.36817905, 0.5789861,  0.12580054, 0.38215966, 0.57599512, 0.1159504,  0.39572824,
            0.57290928, 0.10685038, 0.40881926, 0.56975727, 0.09855521, 0.42148106, 0.56654159,
            0.09104002, 0.43364953, 0.56329296, 0.08434116, 0.44538908, 0.56000859, 0.07841305,
            0.45672421, 0.5566943,  0.07322913, 0.46765017, 0.55336373, 0.06876762, 0.47819138,
            0.5500213,  0.06498436, 0.48839686, 0.54666195, 0.06182163, 0.49828924, 0.5432874,
            0.05922726, 0.50789114, 0.53989827, 0.05714466, 0.51722475, 0.53649429, 0.05551476,
            0.5263115,  0.53307443, 0.05427793, 0.53517186, 0.52963707, 0.05337567, 0.54382515,
            0.52618009, 0.05275208, 0.55228947, 0.52270103, 0.05235479, 0.56058163, 0.51919713,
            0.0521356,  0.56871719, 0.51566545, 0.05205062, 0.57671045, 0.51210292, 0.0520602,
            0.5845745,  0.50850636, 0.05212851, 0.59232129, 0.50487256, 0.05222299, 0.5999617,
            0.50119827, 0.05231367, 0.60750568, 0.49748022, 0.05237234, 0.61496232, 0.49371512,
            0.05237168, 0.62233999, 0.48989963, 0.05228423, 0.62964652, 0.48603032, 0.05208127,
            0.63688935, 0.48210362, 0.05173155, 0.64407572, 0.4781157,  0.0511996,  0.65121289,
            0.47406244, 0.05044367, 0.65830839, 0.46993917, 0.04941288};
  }
}  // namespace